#include "picoled/gfx/2D.h"
#include "picoled/buffer.h"

#include <cmath>

#include <pico/stdio.h>

namespace picoled::gfx
{

triangle triangle::fix_winding() const
{
  if (((b.position - a.position) ^ (c.position - a.position)) < fixed(0))
  {
    return { a, c, b };
  }
  else
  {
    return { a, b, c };
  }
}

vec3 triangle::barycentric(const vec2& p) const
{
  const fixed total = area_double();
  const fixed u = triangle{ b, a, p }.area_double() / total;
  const fixed v = triangle{ a, c, p }.area_double() / total;
  return vec3{ u, v, fixed(1) - u - v };
}

bool triangle::contains(const vec2& p) const
{
  const fixed abp = (b.position - a.position) ^ (p - a.position);
  const fixed bcp = (c.position - b.position) ^ (p - b.position);
  const fixed cap = (a.position - c.position) ^ (p - c.position);

  return abp >= fixed(0) && bcp >= fixed(0) && cap >= fixed(0);
}

vec2 triangle::min() const
{
  return
  {
    fixed::min(fixed::min(a.position.x, b.position.x), c.position.x),
    fixed::min(fixed::min(a.position.y, b.position.y), c.position.y)
  };
}

vec2 triangle::max() const
{
  return
  {
    fixed::max(fixed::max(a.position.x, b.position.x), c.position.x),
    fixed::max(fixed::max(a.position.y, b.position.y), c.position.y)
  };
}

fixed triangle::area_double() const
{
  // don't divide by two since it would cancel out
  // in the barycentric coordinate calculation anyway
  return std::abs((b.position - a.position) ^ (c.position - a.position));
}

void gfx2d::draw_triangle(triangle tri, const buffer<pixel>& texture, bool ignore_winding)
{
  if (ignore_winding) { tri = tri.fix_winding(); }

  const vec2 min = tri.min();
  const vec2 max = tri.max();

  for (fixed i = min.x; i <= max.x; i += 1)
  {
    for (fixed j = min.y; j <= max.y; j += 1)
    {
      const vec2 position = { i, j };
      if (tri.contains(position))
      {
        const vec3 barycentric = tri.barycentric(position);
        const fixed u = barycentric % vec3{ tri.a.texture_coordinates.x, tri.b.texture_coordinates.x, tri.c.texture_coordinates.x };
        const fixed v = barycentric % vec3{ tri.a.texture_coordinates.y, tri.b.texture_coordinates.y, tri.c.texture_coordinates.y };
        const pixel read = texture.read_uv(u, v, buffer_read_mode::clamp);
        screen->write(i, j, read);
      }
    }
  }
}

}
